import yfinance as yf
import numpy as np

def check_optype(op_type):
    if (op_type not in ['p','c']):
        raise ValueError("Input 'p' for put and 'c' for call!")

def check_trtype(tr_type):
    if (tr_type not in ['b','s']):
        raise ValueError("Input 'b' for Buy and 's' for Sell!")  

def payoff_calculator(x, op_type, strike, op_pr, tr_type, n):
    """
    Calculate option payoff at different spot prices.

    Parameters:
    -----------
    x : array-like
        Spot prices
    op_type : str
        'c' for call, 'p' for put
    strike : float
        Strike price
    op_pr : float
        Option premium (per share)
    tr_type : str
        'b' for long, 's' for short
    n : int
        Number of contracts (each contract = 100 shares)

    Returns:
    --------
    array
        Payoff in dollars (accounts for 100 shares per contract)
    """
    y=[]
    if op_type=='c':
        for i in range(len(x)):
            y.append(max((x[i]-strike-op_pr),-op_pr))
    else:
        for i in range(len(x)):
            y.append(max(strike-x[i]-op_pr,-op_pr))
    y=np.array(y)

    if tr_type=='s':
        y=-y
    # Multiply by 100 (shares per contract) and by n (number of contracts)
    return y*n*100

def check_ticker(ticker):
    """
    Check ticker
    """
    try:
        return yf.Ticker('msft').info['currentPrice']
    except KeyError:
        raise ValueError('Ticker not recognized')


import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import yfinance as yf

from .helpers import check_ticker, check_optype, check_trtype, payoff_calculator

import warnings
warnings.filterwarnings('ignore')


abb={'c': 'Call',
    'p': 'Put',
    'b': 'Long',
    's': 'Short'}

def yf_plotter(ticker='msft',exp='default',spot_range=10,
               op_list=[{'op_type':'c','strike':300,'tr_type':'b', 'contract':1},
                        {'op_type':'p','strike':280,'tr_type':'b', 'contract':1}], 
                        save=False, file='fig.png'):
    """
    Plots a basic option payoff diagram for a multiple options and resultant payoff diagram
    
    Parameters
    ----------
    ticker: string, default: 'msft' stock ticker for Microsoft.Inc
       Stock Ticker
    exp: string default: next option expiration date
        Option expiration date in 'YYYY-MM-DD' format
       
    spot_range: int, float, optional, default: 10
       Range of spot variation in percentage 
       
    op_list: list of dictionary
       
       Each dictionary must contiain following keys
       'strike': int, float, default: 720
           Strike Price
       'tr_type': kind {'b', 's'} default:'b'
          Transaction Type>> 'b': long, 's': short
       'op_type': kind {'c','p'}, default:'c'
          Opion type>> 'c': call option, 'p':put option 
       'contracts': int default:1, optional
          Number of contracts
    
    save: Boolean, default False
        Save figure
    
    file: String, default: 'fig.png'
        Filename with extension
    
    Example
    -------
    op1={'op_type':'c','strike':250,'tr_type':'b', 'contract':1}
    op2={'op_type':'p','strike':225,'tr_type':'b','contract':3}
    
    import opstrat as op
    op.yf_plotter(ticker='msft',exp='2021-03-26',spot_range=10, op_list=[op1,op2])
    
    #Plots option payoff diagrams for each op1 and op2 and combined payoff 
    
    """
    #Check input and assigns spot price
    spot=check_ticker(ticker)
    #Expiry dates
    exp_list=yf.Ticker(ticker).options
    
    x=spot*np.arange(100-spot_range,101+spot_range,0.01)/100
    y0=np.zeros_like(x)

    
    def check_exp(exp):
        """
        Check expiry date
        """
        if exp not in exp_list:
            raise ValueError('Option for the given date not available!')
    
    if exp=='default':
        exp=yf.Ticker('msft').options[0]
    else:
        check_exp(exp)
               
    def check_strike(df, strike):
        if strike not in df.strike.unique():
            raise ValueError('Option for the given Strike Price not available!')
    
    y_list=[]
    
    for op in op_list:
        op_type=str.lower(op['op_type'])
        tr_type=str.lower(op['tr_type'])
        
        check_optype(op_type)
        check_trtype(tr_type)
    
        if(op_type=='p'):
            df=yf.Ticker(ticker).option_chain(exp).puts
        else:
            df=yf.Ticker(ticker).option_chain(exp).calls
    
        strike=op['strike']
        check_strike(df, strike)
        op_pr=df[df.strike==strike].lastPrice.sum()
        try:
            contract=op['contract']
        except:
            contract=1
        
        y_list.append(payoff_calculator(x, op_type, strike, op_pr, tr_type, contract))
    

    def plotter():
        y=0
        plt.figure(figsize=(10,6))
        for i in range (len(op_list)):
            try:
                contract=str(op_list[i]['contract'])  
            except:
                contract='1'
                
            label=contract+' '+str(abb[op_list[i]['tr_type']])+' '+str(abb[op_list[i]['op_type']])+' ST: '+str(op_list[i]['strike'])
            sns.lineplot(x=x, y=y_list[i], label=label, alpha=0.5)
            y+=np.array(y_list[i])
        
        sns.lineplot(x=x, y=y, label='combined', alpha=1, color='k')
        plt.axhline(color='k', linestyle='--')
        plt.axvline(x=spot, color='r', linestyle='--', label='spot price')
        plt.legend()
        plt.legend(loc='upper right')
        title="OPTION STRATEGY ("+str.upper(ticker)+') '+' Exp :'+str(exp)
        plt.fill_between(x, y, 0, alpha=0.2, where=y>y0, facecolor='green', interpolate=True)
        plt.fill_between(x, y, 0, alpha=0.2, where=y<y0, facecolor='red', interpolate=True)
        plt.title(title)
        plt.tight_layout()
        if save==True:
            plt.savefig(file)
        plt.show()

    plotter()           


import numpy as np
from scipy.stats import norm


def black_scholes(t=40, r=4.00, v=32.00, K=60, St=62, type='c'):
    """
    Parameters:
    K : Exercise Price (Strike)
    St: Current Stock Price (Spot)
    v : Volatility in percentage
    r : Risk free rate in percentage
    t : Time to expiration in days
    type: Type of option 'c' for call 'p' for put
    default: 'c'
    """
    # if type = 'c' or 'C' call option else put option
    try:
        type=type.lower()
        if(type=='c'):
            option_type='call'
        else:
            option_type='put'
    except:
        option_type='put'

    #Check time 
    try:
        #convert time in days to years
        t=t/365
    except:
        raise TypeError("Enter numerical value for time")

    #Check risk free rate 
    try:
        #convert percentage to decimal
        r=r/100
    except:
        raise TypeError("Enter numerical value for risk free rate")
    
    #Check volatility
    try:
        #convert percentage to decimal
        v=v/100
    except:
        raise TypeError("Enter numerical value for volatility")  

    #Check Stock Price
    try:
        St=St+0
    except:
        raise TypeError("Enter numerical value for stock price")
    
    #Check Exercise Price
    try:
        K=K+0
    except:
        raise TypeError("Enter numerical value for Exercise price")    
    
    n1=np.log(St/K)
    n2=(r+(np.power(v,2)/2))*t
    d=v*(np.sqrt(t))

    d1=(n1+n2)/d
    d2=d1-(v*np.sqrt(t))

    if type=='c':
        N_d1=norm.cdf(d1)
        N_d2=norm.cdf(d2)
    else:
        N_d1=norm.cdf(-d1)
        N_d2=norm.cdf(-d2)

    A=(St*N_d1)
    B=(K*N_d2*(np.exp(-r*t)))

    if type=='c':
        val=A-B
        val_int=max(0,St-K)
    else:
        val=B-A
        val_int=max(0,K-St)
    val_time=val-val_int

    # Option values in dictionary
    value={'option value':val, 'intrinsic value':val_int, 'time value':val_time}

    #CALCULATE OPTION GREEKS
    if type=='c':
        delta=N_d1
        theta=(-((St*v*np.exp(-np.power(d1,2)/2))/(np.sqrt(8*np.pi*t)))-(N_d2*r*K*np.exp(-r*t)))/365
        rho=t*K*N_d2*np.exp(-r*t)/100
    else:
        delta=-N_d1
        theta=(-((St*v*np.exp(-np.power(d1,2)/2))/(np.sqrt(8*np.pi*t)))+(N_d2*r*K*np.exp(-r*t)))/365
        rho=-t*K*N_d2*np.exp(-r*t)/100

    gamma=(np.exp(-np.power(d1,2)/2))/(St*v*np.sqrt(2*np.pi*t))
    vega=(St*np.sqrt(t)*np.exp(-np.power(d1,2)/2))/(np.sqrt(2*np.pi)*100)
    
    #Option greeks in Dictionary
    greeks={'delta':delta, 'gamma':gamma, 'theta':theta, 'vega':vega, 'rho':rho}

    return {'value':value, 'greeks':greeks}


#multiplotter
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

from .helpers import payoff_calculator, check_optype, check_trtype
from .blackscholes import black_scholes

abb={'c': 'Call',
    'p': 'Put',
    'b': 'Long',
    's': 'Short'}

def multi_plotter(spot_range=20, spot=100,
                op_list=[{'op_type':'c','strike':110,'tr_type':'s','op_pr':2,'contract':1},
                {'op_type':'p','strike':95,'tr_type':'s','op_pr':6,'contract':1}],
                  save=False, file='fig.png', risk_free_rate=4.00, default_volatility=32.00, y_limiter=None):
    """
    Plots a basic option payoff diagram for a multiple options and resultant payoff diagram

    Parameters
    ----------
    spot: int, float, default: 100
       Spot Price

    spot_range: int, float, optional, default: 20
       Range of spot variation in percentage

    op_list: list of dictionary

       Each dictionary must contiain following keys
       'strike': int, float, default: 720
           Strike Price
       'tr_type': kind {'b', 's'} default:'b'
          Transaction Type>> 'b': long, 's': short
       'op_pr': int, float, default: 10
          Option Price (required if 'dte' is not provided)
       'op_type': kind {'c','p'}, default:'c'
          Opion type>> 'c': call option, 'p':put option
       'contracts': int default:1, optional
           Number of contracts
       'dte': int, optional
           Days to expiration. If provided for one option, must be provided for all.
           When provided, uses Black-Scholes model to plot daily option values instead of payoff diagrams.
       'volatility': float, optional
           Implied volatility in percentage for this specific option leg.
           If not provided, uses the default_volatility parameter.

    save: Boolean, default False
        Save figure

    file: String, default: 'fig.png'
        Filename with extension

    risk_free_rate: float, default: 4.00
        Risk free rate in percentage

    default_volatility: float, default: 32.00
        Default volatility in percentage used for all option legs that don't specify their own volatility.
        Only used when 'dte' is provided for options.

    y_limiter: callable, optional, default: None
        A function that takes the combined payoff array (y) and returns a tuple (y_min, y_max) for setting ylim.
        If provided, will be called to set the y-axis limits. If None, matplotlib auto-scales the y-axis.
        Example: lambda y: (-max(abs(y)), max(abs(y))) for symmetric limits

    Example
    -------
    op1={'op_type':'c','strike':110,'tr_type':'s','op_pr':2,'contract':1}
    op2={'op_type':'p','strike':95,'tr_type':'s','op_pr':6,'contract':1}

    import superOpStrat  as op
    op.multi_plotter(spot_range=20, spot=100, op_list=[op1,op2])

    #Plots option payoff diagrams for each op1 and op2 and combined payoff

    Example with DTE and per-leg volatility
    ----------------------------------------
    op1={'op_type':'c','strike':110,'tr_type':'b','dte':30,'contract':1,'volatility':25.0}
    op2={'op_type':'p','strike':95,'tr_type':'b','dte':30,'contract':1,'volatility':28.5}

    op.multi_plotter(spot_range=20, spot=100, op_list=[op1,op2], default_volatility=32.00)

    #Plots option values using Black-Scholes with different volatilities for each leg

    """
    x=spot*np.arange(100-spot_range,101+spot_range,0.01)/100
    y0=np.zeros_like(x)

    # Check if DTE is provided and validate consistency
    has_dte = any('dte' in op for op in op_list)
    if has_dte:
        # If any option has DTE, all must have DTE
        if not all('dte' in op for op in op_list):
            raise ValueError("If 'dte' is provided for one option, it must be provided for all options")

    y_list=[]
    for op in op_list:
        op_type=str.lower(op['op_type'])
        tr_type=str.lower(op['tr_type'])
        check_optype(op_type)
        check_trtype(tr_type)

        strike=op['strike']
        try:
            contract=op['contract']
        except:
            contract=1

        if has_dte:
            # Use Black-Scholes to calculate daily option values
            dte=op['dte']
            # Use per-leg volatility if provided, otherwise use default
            volatility=op.get('volatility', default_volatility)
            y_values=[]
            for spot_price in x:
                # Call black_scholes with per-leg volatility
                bs_result=black_scholes(t=dte, r=risk_free_rate, v=volatility, K=strike, St=spot_price, type=op_type)
                option_value=bs_result['value']['option value']

                # Apply transaction type (short = negative)
                if tr_type=='s':
                    option_value=-option_value

                # Apply contract multiplier
                option_value=option_value*contract

                y_values.append(option_value)
            y_list.append(np.array(y_values))
        else:
            # Use original payoff calculator (requires op_pr)
            op_pr=op['op_pr']
            y_list.append(payoff_calculator(x, op_type, strike, op_pr, tr_type, contract))
    

    def plotter():
        y=0
        plt.figure(figsize=(10,6))
        for i in range (len(op_list)):
            try:
                contract=str(op_list[i]['contract'])  
            except:
                contract='1'
                
            label=contract+' '+str(abb[op_list[i]['tr_type']])+' '+str(abb[op_list[i]['op_type']])+' ST: '+str(op_list[i]['strike'])
            sns.lineplot(x=x, y=y_list[i], label=label, alpha=0.5)
            y+=np.array(y_list[i])
        
        sns.lineplot(x=x, y=y, label='combined', alpha=1, color='k')
        plt.axhline(color='k', linestyle='--')
        plt.axvline(x=spot, color='r', linestyle='--', label='spot price')
        plt.legend()
        plt.legend(loc='upper right')
        title="Multiple Options Plotter"
        plt.title(title)
        plt.fill_between(x, y, 0, alpha=0.2, where=y>y0, facecolor='green', interpolate=True)
        plt.fill_between(x, y, 0, alpha=0.2, where=y<y0, facecolor='red', interpolate=True)

        if y_limiter is not None:
            y_limits = y_limiter(y)
            plt.ylim(y_limits)

        plt.tight_layout()
        if save==True:
            plt.savefig(file)
        plt.show()

    plotter()      


from codecs import open
from setuptools import setup, find_packages

with open('README.md') as readme_file:
    README = readme_file.read()


VERSION = '0.1.2'
DESCRIPTION = 'Enhanced options payoff diagram charting based on opstrat library'
URL = 'https://github.com/pursual/superOpStrat'

# Setting up
setup(
    name="superOpStrat",
    version=VERSION,
    author="Tony G",
    author_email="<pursual@gmail.com>",
    description=DESCRIPTION,
    long_description_content_type='text/markdown',
    long_description=README,
    url=URL,
    license='MIT',
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    python_requires='>=3.6',
    install_requires=['matplotlib', 
                      'pandas', 
                      'numpy',
                      'scipy',
                      'seaborn',
                      'yfinance'],
    keywords=['python', 
              'options', 
              'finance', 
              'opstrat', 
              'data visualization', 
              'stock market'],
    classifiers=[
            "Development Status :: 4 - Beta",
            "Intended Audience :: Developers",
            "Intended Audience :: Education",
            "Intended Audience :: Science/Research",
            "Intended Audience :: Financial and Insurance Industry",
            "License :: OSI Approved :: MIT License",
            "Natural Language :: English",
            "Operating System :: OS Independent",
            "Topic :: Scientific/Engineering :: Visualization",
            'Topic :: Scientific/Engineering :: Information Analysis',
            "Topic :: Office/Business :: Financial :: Investment",
            "Topic :: Office/Business :: Financial",
            "Programming Language :: Python :: 3"]
)
